//
//	Fiview -- Filter design/viewing/comparison tool
//
//        Copyright (c) 2002-2003 Jim Peters <http://uazu.net/>.
//        Released under the GNU GPL version 2 as published by the
//        Free Software Foundation.  See the file COPYING for details,
//        or visit <http://www.gnu.org/copyleft/gpl.html>.
//

#include "all.h"

//
//	Globals
//

// Display-related
SDL_Surface *disp;	// Display
Uint32 *disp_pix32;	// Pixel data for 32-bit display, else 0
Uint16 *disp_pix16;	// Pixel data for 16-bit display, else 0
int disp_my;            // Display pitch, in pixels
int disp_sx, disp_sy;   // Display size

int disp_rl, disp_rs;	// Red component transformation
int disp_gl, disp_gs;	// Green component transformation
int disp_bl, disp_bs;	// Blue component transformation
int disp_am;		// Alpha mask

int font_sx, font_sy;	// Default font size
int *colour;		// Array of colours mapped to current display: see colour_data

// Display areas
Rect d_info;		// Top-left info display
Rect d_freq;		// Mini frequency display
Rect d_time;		// Mini time display
Rect d_minf;		// Main area info line
Rect d_main;		// Main display area

// Temporary buffer for display routines, each disp_sx entries long
double *d_wrk1;

// Arguments
double a_f0, a_f1;	// Frequency range given by -f command-line arguments
int a_adj;		// Automatically adjust frequencies to meet 50% levels

// Settings
Filter *curr;		// Currently displayed filter
double s_rate;		// Base sampling rate, or 0 if not set
int s_main;		// Main display type: 'F' freq, 'T' time, 'I' info, 'H' help

int s_tmzoom;		// Time mini-display zoom (samples/pixel if +ve, pixels/sample if -ve)
int s_zoom;		// Special x10 zoom for freq + time displays

double s_freq0;		// Frequency range for main frequency display
double s_freq1;		//  (if shown)

double s_tim0;		// Time range for main display (if shown)
double s_tim1;		//  tim0 -> tim1  (these are sample-counts, <= 0)

int s_pager_lin;	// Current top-line in pager
char *s_pager_txt;	// Current pager text
char *s_pager_inf;	// Info line for pager
int s_pager_cnt;	// Number of lines in current text
int s_pager_at_end;	// Reached end of pager text
int s_pager_typ;	// s_main type that pager reflects

// Currently buffered time-display
double c_tim0;
double c_tim1;
int c_tim_sx;
Filter *c_tim_filt;
float *c_tim_buf;

// Currently buffered freq-display
double c_freq0;
double c_freq1;
int c_freq_sx;
Filter *c_freq_filt;
double *c_freq_buf;

// Currently buffered info-display
Filter *c_info_filt;

// Main loop
int rearrange;		// Set if a rearrange of the screen is necessary
int redraw;		// Set to request a redraw of the screen
int part_cmd= 0;	// Partial command status, or 0
int n_filt= 0;		// Number of filters loaded

// Hacks
double nan_global;	// Used on MSVC because 0.0/0.0 as a constant is not understood

// Misc
char *helptext;		// Generated help-text


//
//      Utility functions
//

void 
error(char *fmt, ...) {
   va_list ap;
   va_start(ap, fmt);
   fprintf(stderr, PROGNAME ": ");
   vfprintf(stderr, fmt, ap);
   fprintf(stderr, "\n");
   exit(1);
}

void
errorSDL(char *fmt, ...) {
   va_list ap;
   va_start(ap, fmt);
   fprintf(stderr, PROGNAME ": ");
   vfprintf(stderr, fmt, ap);
   fprintf(stderr, "\n  %s\n", SDL_GetError());
   exit(1);
}

#define NL "\n"

void
usage() {
   error("Filter Design, Viewing and Comparison tool, version " VERSION
	 NL "Copyright (c) 2002-2003 Jim Peters, http://uazu.net/, all rights "
	 NL "  reserved, released under the GNU GPL v2.  See file COPYING."
	 NL "LibSDL code from the SDL project (http://libsdl.org/) is released under"
	 NL "  the GNU LGPL version 2 or later.  See file COPYING.LIB."
	 NL "Filter design code derived from the mkfilter package by Tony Fisher"
	 NL "  See here: <http://www-users.cs.york.ac.uk/~fisher/mkfilter/>"
	 NL
	 NL "Usage: fiview [options] [<sample-rate>] [<filenames>] [-i <spec-words> ...]"
	 NL
	 NL "Options:"
	 NL "  -F <mode>     Run full-screen with the given mode, <wid>x<hgt>x<bpp>"
	 NL "                <bpp> may be 16 or 32.  For example: 800x600x16"
	 NL "  -W <size>     Run as a window with the given size: <wid>x<hgt>"
	 NL "  -f <freq>[-<freq>]    Specify the default frequency or frequency range for"
	 NL "                        any filter specifications used."
	 NL "  -a            Automatically adjust the frequencies fed to the filter design"
	 NL "                  code to give accurate 50%% levels at the given frequencies"
	 NL "  -h            Dump help-text to STDOUT and exit"
	 NL ""
	 NL "Filter specifications are read from all the filenames given, and also from the"
	 NL "remainder of the command line if the -i (immediate) option is given."
	 );
}

void warn(char *fmt, ...) {
   va_list ap;
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   fprintf(stderr, "\n");
}

void *Alloc(int size) {
   void *vp= calloc(1, size);
   if (!vp) error("Out of memory");
   return vp;
}

void *StrDup(char *str) {
   char *cp= strdup(str);
   if (!cp) error("Out of memory");
   return cp;
}


//
//	Main routine
//

int 
main(int ac, char **av) {
   char dmy;
   int sx= 800, sy= 600, bpp= 0;	// Default is 800x600 resizable window
   SDL_Event ev;
   int motion_xx, motion_yy;
   int immed= 0;

   // Generate helptext
   helptext= gen_helptext();

   // Process arguments
   a_f0= a_f1= -1.0;		// Undefined
   ac--; av++;
   while (ac > 0 && av[0][0] == '-' && av[0][1]) {
      char ch, *p= *av++ + 1; 
      ac--;
      while ((ch= *p++)) switch (ch) {
       case 'F':
	  if (ac-- < 1) usage();
	  if (3 != sscanf(*av++, "%dx%dx%d %c", &sx, &sy, &bpp, &dmy) ||
	      (bpp != 16 && bpp != 32))
	     error("Bad mode-spec: %s", av[-1]);
	  break;
       case 'W':
	  if (ac-- < 1) usage();
	  if (2 != sscanf(*av++, "%dx%d %c", &sx, &sy, &dmy))
	     error("Bad window size: %s", av[-1]);
	  break;
       case 'a':
	  error("@@@ Auto-adjust not yet implemented");
	  a_adj= 1;
	  break;
       case 'f':
	  if (ac-- < 1) usage();
	  if (1 == sscanf(av[0], "%lf %c", &a_f0, &dmy)) {
	     a_f1= a_f0;
	  } else if (2 == sscanf(av[0], "%lf-%lf %c", &a_f0, &a_f1, &dmy)) {
	     ;
	  } else 
	     error("Bad frequency or frequency range: %s", av[0]);
	  av++;
	  break;
       case 'h':
	  printf("%s", helptext);
	  exit(0);
       case 'i':
	  immed= 1;
	  break;
       default:	
	  error("Unknown option '%c'", ch);
      }
   }

   // See if we have a sampling rate provided
   s_rate= 0;
   if (!immed) {
      if (ac < 1) usage();
      if (1 == sscanf(av[0], "%lf %c", &s_rate, &dmy)) {
	 ac--; av++;
      } else {
	 s_rate= 0;
      }
   }

   // Load up all the filter files
   if (!immed) {
      while (ac && 0 != strcmp("-i", av[0])) {
	 n_filt= filter_load_file(*av++);
	 ac--;
      }
      if (ac) { immed= 1; ac--; av++; }
   }

   // Load up any immediate filter specified on the command line
   if (immed) {
      char *txt;
      scr_zap();
      while (ac > 0) { scr_pr("%s ", *av++); ac--; }
      txt= scr_dup();
      filter_load_immed(txt);
      free(txt);
   }

   // Dump all the loaded filters
   {
      FILE *log_fp;
      Filter *ff;
      int a;

      if (!(log_fp= fopen("fiview.log", "w")))
	 error("Can't create file: fiview.log");

      for (a= 1; (ff= filter_find(a)); a++) {
	 fputs("// Generated by Fiview " VERSION 
	       " <http://uazu.net/fiview/>.  \n", log_fp);
	 fputs("// All generated example code below is in the public domain.\n", log_fp);
	 fputs(ff->dump, log_fp);
      }
      
      fputs(filter_standard_code(), log_fp);
      if (fclose(log_fp))
	 error("Error writing file: fiview.log");
   }

   // Initialize SDL
   if (0 > SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE))
      errorSDL("Couldn't initialize SDL");
   atexit(SDL_Quit);
   //SDL_EnableUNICODE(1);		// So we get translated keypresses
   SDL_EnableKeyRepeat(200, 40);

   // Initialise graphics
   graphics_init(sx, sy, bpp);
   font_sy= 16;
   font_sx= 8;
   curr= filter_find(1);
   s_main= 'H';
   rearrange= 1;

   // Main loop
   status("+");
   while (1) {
      if (rearrange) {
	 arrange_display();
	 set_s_tmzoom();
	 auto_adjust_time();
	 auto_adjust_freq();
	 rearrange= 0;
	 redraw= 1;
      }
      if (redraw) {
	 // Suspend update during redraw so that we get a clean change
	 suspend_update= 1;
	 draw_mini_info();
	 draw_mini_freq();
	 draw_mini_time();
	 switch (s_main) {
	  case 'T': draw_time(); break;
	  case 'F': draw_freq(); break;
	  case 'I': draw_info(); break;
	  case 'H': draw_help(); break;
	 }
	 draw_status();
	 suspend_update= 0;
	 update_all();
	 redraw= 0;
      }

      // Wait for an event
      if (!SDL_WaitEvent(0)) 
	 errorSDL("Unexpected error waiting for events");

      // Process all outstanding events
      motion_yy= motion_xx= -1;
      while (SDL_PollEvent(&ev)) {
	 int key;
	 switch (ev.type) {
	  case SDL_KEYUP:
	     if (s_zoom) {
		s_zoom= 0;
		redraw= 1;
		continue;
	     }
	     break;
	  case SDL_KEYDOWN:
	     status("");	// Use keypress to clear temporary messages from status line
	     key= ev.key.keysym.sym;
	     //With translation only: key= ev.key.keysym.unicode;

	     // Switching between filters
	     if (key >= '1' && key <= '9') {
		Filter *ff= filter_find(key - '0');
		if (ff) { curr= ff; redraw= 1; }
		continue;
	     }

	     // General keys
	     switch (key) {
	      case SDLK_F1:
	      case 'h':
		 s_main= 'H';
		 redraw= 1;
		 continue;
	      case SDLK_F2:
	      case 'i':
		 s_main= 'I';
		 redraw= 1;
		 continue;
	      case SDLK_F3:
	      case 'f':
		 s_main= 'F';
		 redraw= 1;
		 continue;
	      case SDLK_F4:
	      case 't':
		 s_main= 'T';
		 redraw= 1;
		 continue;
	      case SDLK_ESCAPE:
	      case 'q':
		 exit(0);
	     }

	     switch (s_main) {
	      case 'F':		// Frequency display handling
		 switch (key) {
		    double adj, wid, mid;
		  case 'a': 
		     auto_adjust_freq(); 
		     draw_freq(); draw_mini_freq();
		     continue;
		  case SDLK_LEFT:
		     adj= 0.25 * (s_freq1-s_freq0);
		     s_freq0 -= adj; s_freq1 -= adj;
		     draw_freq(); draw_mini_freq();
		     continue;
		  case SDLK_RIGHT:
		     adj= 0.25 * (s_freq1-s_freq0);
		     s_freq0 += adj; s_freq1 += adj;
		     draw_freq(); draw_mini_freq();
		     continue;
		  case SDLK_UP:
		     wid= s_freq1-s_freq0;
		     mid= (s_freq0 + s_freq1) * 0.5;
		     s_freq0= mid-wid; s_freq1= mid+wid;
		     draw_freq(); draw_mini_freq();
		     continue;
		  case SDLK_DOWN:
		     wid= (s_freq1-s_freq0) / 4.0;
		     mid= (s_freq0 + s_freq1) * 0.5;
		     s_freq0= mid-wid; s_freq1= mid+wid;
		     draw_freq(); draw_mini_freq();
		     continue;
		  case SDLK_BACKSPACE:
		  case SDLK_PAGEUP:
		     adj= (s_freq1-s_freq0);
		     s_freq0 -= adj; s_freq1 -= adj;
		     draw_freq(); draw_mini_freq();
		     continue;
		  case SDLK_SPACE:
		  case SDLK_PAGEDOWN:
		     adj= (s_freq1-s_freq0);
		     s_freq0 += adj; s_freq1 += adj;
		     draw_freq(); draw_mini_freq();
		     continue;
		  case 'z':
		     if (!s_zoom) {
			s_zoom= 1;
			draw_freq();
		     }
		     continue;
		 }
		 break;
	      case 'T':		// Time display handling
		 switch (key) {
		    double adj, mid, wid;
		  case 'a': 
		     auto_adjust_time(); 
		     draw_time(); draw_mini_time(); 
		     continue;
		  case SDLK_LEFT:
		     adj= (s_tim1 - s_tim0) * 0.25;
		     s_tim0 -= adj; s_tim1 -= adj;
		     draw_time(); draw_mini_time(); 
		     continue;
		  case SDLK_RIGHT:
		     adj= (s_tim1 - s_tim0) * 0.25;
		     s_tim0 += adj; s_tim1 += adj;
		     draw_time(); draw_mini_time(); 
		     continue;
		  case SDLK_UP:
		     wid= s_tim1-s_tim0;
		     mid= (s_tim1 + s_tim0) * 0.5;
		     s_tim0= mid-wid; s_tim1= mid+wid;
		     draw_time(); draw_mini_time(); 
		     continue;
		  case SDLK_DOWN:
		     wid= (s_tim1-s_tim0) * 0.25;
		     mid= (s_tim1 + s_tim0) * 0.5;
		     s_tim0= mid-wid; s_tim1= mid+wid;
		     draw_time(); draw_mini_time(); 
		     continue;
		  case SDLK_BACKSPACE:
		  case SDLK_PAGEUP:
		     adj= (s_tim1 - s_tim0);
		     s_tim0 -= adj; s_tim1 -= adj;
		     draw_time(); draw_mini_time(); 
		     continue;
		  case SDLK_SPACE:
		  case SDLK_PAGEDOWN:
		     adj= (s_tim1 - s_tim0);
		     s_tim0 += adj; s_tim1 += adj;
		     draw_time(); draw_mini_time(); 
		     continue;
		  case 'z':
		     if (!s_zoom) {
			s_zoom= 1;
			draw_time();
		     }
		     continue;
		 }
		 break;
	      case 'I':		// Pager display handling
	      case 'H':
		 switch (key) {
		  case SDLK_UP:
		     if (s_pager_lin) {
			s_pager_lin--;
			draw_pager();
		     }
		     continue;
		  case SDLK_RETURN:
		  case SDLK_DOWN:
		     if (!s_pager_at_end) {
			s_pager_lin++;
			draw_pager();
		     }
		     continue;
		  case SDLK_BACKSPACE:
		  case SDLK_PAGEUP:
		     if (s_pager_lin) {
			s_pager_lin -= d_main.sy/font_sy - 2;
			if (s_pager_lin < 0) s_pager_lin= 0;
			draw_pager();
		     }
		     continue;
		  case SDLK_SPACE:
		  case SDLK_PAGEDOWN:
		     if (!s_pager_at_end) {
			s_pager_lin += (d_main.sy/font_sy - 2);
			draw_pager();
		     }
		     continue;
		  case 'd':
		     if (!s_pager_at_end) {
			s_pager_lin += (d_main.sy/font_sy - 2) / 2;
			draw_pager();
		     }
		     continue;
		 }
		 break;
	     }
	     break;
	  case SDL_MOUSEMOTION:
	     // Save just the last motion until the end of the loop to
	     // avoid incredible slowdown, because some of the
	     // calculations might be too heavy for a stream of motion
	     // events.
	     motion_xx= ev.motion.x;
	     motion_yy= ev.motion.y;
	     continue;
	  case SDL_MOUSEBUTTONDOWN:
	     if (inRect(&d_info, ev.button.x, ev.button.y) && s_main != 'I') {
		s_main= 'I'; redraw= 1; continue;
	     }
	     if (inRect(&d_freq, ev.button.x, ev.button.y) && s_main != 'F') {
		s_main= 'F'; redraw= 1; continue;
	     }
	     if (inRect(&d_time, ev.button.x, ev.button.y) && s_main != 'T') {
		s_main= 'T'; redraw= 1; continue;
	     }
	     continue;
	  case SDL_VIDEORESIZE:
	     sx= ev.resize.w;
	     sy= ev.resize.h;
	     graphics_init(sx, sy, 0);
	     rearrange= 1;
	     continue;
	  case SDL_QUIT:
	     exit(0);
	 }
      }

      // Handle motion events now
      if (motion_yy >= d_main.yy &&
	  motion_yy - d_main.yy < d_main.sy) {
	 switch (s_main) {
	  case 'F':
	     show_freq_status(motion_xx - d_main.xx, 
			      motion_yy - d_main.yy);    
	     break;
	  case 'T':
	     show_time_status(motion_xx - d_main.xx, 
			      motion_yy - d_main.yy);    
	     break;
	 }
      }

      // Loop ...
   }

   return 0;
}

//
//	Check to see if a point is within the given rectangle
//

int 
inRect(Rect *rr, int xx, int yy) {
   return (xx >= rr->xx &&
	   xx < rr->xx + rr->sx &&
	   yy >= rr->yy &&
	   yy < rr->yy + rr->sy);
}

//
//	Automatically adjust filter viewing settings to show most of
//	what is interesting.
//

void 
auto_adjust_time() {
   // Adjust time display to show up to 95% point
   s_tim0= -d_main.sx;
   s_tim1= 0;

   if (curr->cnt95 > 0) {
      while (-curr->cnt95 < s_tim0) 
	 s_tim0 *= 2;
      while (-curr->cnt95 > s_tim0*0.5) 
	 s_tim0 *= 0.5;
   }
}

void 
auto_adjust_freq() {
   // Adjust frequency display to show all the interesting 50% points
   s_freq0= 0;
   s_freq1= 0.5;
   
   if (curr->n_m3db) {
      int i0= 0;
      int i1= curr->n_m3db-1;
      double mid;
      double wid;
      double f0, f1;

      if (i1-i0 >= 3 && curr->m3db[i0] == 0 && curr->m3db[i1] == 0.5) {
	 // Don't include start and end if both are present, because
	 // this could be a band-cut filter, in which case we need to
	 // focus on the cut.
	 i0++; i1--;
      }

      f0= curr->m3db[i0];
      f1= curr->m3db[i1];

      if (i0+1 == i1 && (f0 != 0.0 || f1 != 0.5)) {
	 // Adjust to most interesting end of range
	 if (f0 == 0.0 && f1 > 0.25) {
	    f0= f1; f1= 0.5;
	 } else if (f1 == 0.5 && f0 < 0.25) {
	    f1= f0; f0= 0.0;
	 }
      }
      
      mid= 0.5 * (f0 + f1);
      wid= (f1-f0) * 1.7;		// Allow 70% extra space
      f0= mid-wid; 
      f1= mid+wid;
      if (f0 < 0.0) f0= 0.0;
      if (f1 > 0.5) f1= 0.5;

      // Start at 0.0-0.5 view
      s_freq0= 0.0;
      s_freq1= 0.5;

      // Zoom in stages, emulating user actions
      while (1) {
	 double tf0, tf1, inc;	// Trial frequency range
	 tf0= s_freq0;
	 tf1= (s_freq0+s_freq1) * 0.5;
	 inc= (tf1-tf0) * 0.25;
	 while (tf0 + inc < f0) { tf0 += inc; tf1 += inc; }
	 if (tf1 < f1) break;		// Doesn't fit
	 s_freq0= tf0;
	 s_freq1= tf1;
      }
   }
}

//
//	Find the zoom level required for the mini time-display in
//	order to show at least 99% of the impulse response of all the
//	loaded filters.
//

void 
set_s_tmzoom() {
   Filter *ff;
   s_tmzoom= -d_time.sx;

   for (ff= filters; ff; ff= ff->nxt) {
      int tmz= 0;
      if (d_time.sx >= 2 * (ff->cnt99+1))
	 tmz= -(d_time.sx / (ff->cnt99+1));
      else 
	 tmz= (ff->cnt99 - 1) / d_time.sx + 1;
      if (s_tmzoom < tmz) s_tmzoom= tmz;
   }

   // For infinite filters, 99% point is 0, so display one per pixel
   if (s_tmzoom == -d_time.sx) s_tmzoom= 1;
}
   

// END //
